import pandas as pd
import logging

from fhir2dataset.graphquery import GraphQuery
from fhir2dataset.fhirrules_getter import FHIRRules
from fhir2dataset.api_caller import ApiGetter
from fhir2dataset.url_builder import URLBuilder
from fhir2dataset.graph_tools import join_path
from fhir2dataset.timer import timing


logger = logging.getLogger(__name__)


class Query:
    """Query Executor

    Description:
        This module purpose is to perform SQL-type queries whose information is 
        filled in configuration files of the form:
        ```
        {
            "from": {
                "alias n°1": "Resource type 1",
                "alias n°2": "Resource type 2",
                "alias n°3": "Resource type 3",
                ...
            },
            "select": {
                "alias n°1": [
                    "searchparam/fhirpath of attribute a of resource type 1",
                    "searchparam/fhirpath of attribute b of resource type 1",
                    "searchparam/fhirpath of attribute c of resource type 1"
                ],
                "alias n°2": [
                    "searchparam of attribute a of resource type 2",
                    ...
                ],
                ...
            },
            "join": {
                "inner": {
                    "alias n°1": {
                        "searchparam of attribute d, which is of type Reference, of resource type 1": "alias n°2"
                    },
                    ...
                },
                "child": {
                    "alias n°2": {
                        "searchparam of attribute b, which is of type Reference, of resource type 2": "alias n°3"
                    },
                    ...
                },
                "parent": {
                    ...
                }
            },
            "where": {
                "alias n°2": {
                    "searchparam of attribute c of resource type 2": "value 1",
                    "searchparam of attribute d of resource type 2": "value 2"
                },
                "alias n°3": {
                    "searchparam of attribute a of resource type 2": "value 3",
                    "searchparam of attribute b of resource type 2": "value 4"
                },
                ...
            }
        }
        ```
        for the next associated SQL query:
        ```
        SELECT (alias n°1).a, (alias n°1).b, (alias n°1).c, (alias n°2).a FROM (Resource type 1) as (alias n°1)
        INNER JOIN (Resource type 2) as (alias n°2) 
        ON (alias n°1).d = (alias n°2) 
        INNER JOIN (Resource type 3) as (alias n°3) 
        ON (alias n°2).b = (alias n°3) WHERE (alias n°2).c = "value 1"  
        AND (alias n°2).d = "value 2"  
        AND (alias n°3).a = "value 3"  
        AND (alias n°3).b = "value 4"
    ```
    
    Attributes:
        config {dict} -- dictionary storing the initial request
        graph_query {type(GraphQuery)} -- instance of a GraphQuery object that gives a graphical representation of the query
        dataframes {dict} -- dictionary storing for each alias the resources requested on the api in tabular format
        main_dataframe {DataFrame} -- pandas dataframe storing the final result table
    
    Example:
        import fhir2dataset
        import logging

        logging.basicConfig(level=logging.INFO)
        query = Query(fhir_api_url=fhir_api_url, fhir_rules=fhir_rules, token=token)
        query.from_config(config)
        query.execute()
        df = query.main_dataframe
    """  # noqa

    @timing
    def __init__(self, fhir_api_url: str, fhir_rules: type(FHIRRules) = None, token: str = None):
        """Requestor's initialisation 

        Arguments:
            fhir_api_url {str} -- The Service Base URL (e.g. http://hapi.fhir.org/baseR4/)

        Keyword Arguments:
            fhir_rules {type(FHIRRules)} -- an instance of a FHIRRules-type object. If the instance is not filled a default version will be used. (default: {None})
            token {str} -- bearer token authentication if necessary (default: {None})
        """  # noqa
        self.fhir_api_url = fhir_api_url
        if not fhir_rules:
            fhir_rules = FHIRRules(fhir_api_url=self.fhir_api_url)
        self.fhir_rules = fhir_rules
        self.token = token

        self.config = None
        self.graph_query = None
        self.dataframes = dict()
        self.main_dataframe = None

    @timing
    def from_config(self, config: dict):
        """Executes the query from a dictionary in the format of a configuration file

        Arguments:
            config {dict} -- dictionary in the format of a configuration file
        """
        self.config = {
            "from_dict": config.get("from", None),
            "select_dict": config.get("select", None),
            "where_dict": config.get("where", None),
            "join_dict": config.get("join", None),
        }

    @timing
    def execute(self, debug: bool = False):
        """Executes the complete query

        1. constructs a GraphQuery object to store the query as a graph
        2. builds the url of the requests to send to the API
        3. retrieves the answers from the api and puts them as tables in the dataframes attribute
        4. executes joins to return the result table in the main_dataframe attribute

        Keyword Arguments::
            debug {bool} -- if debug is true then the columns needed for internal processing are kept in the final dataframe. Otherwise only the columns of the select are kept in the final dataframe. (default: {False})
        """  # noqa
        self.graph_query = GraphQuery(fhir_api_url=self.fhir_api_url, fhir_rules=self.fhir_rules)
        self.graph_query.execute(**self.config)
        for resource_alias in self.graph_query.resources_alias_info.keys():
            resource_alias_info = self.graph_query.resources_alias_info[resource_alias]
            elements = resource_alias_info.elements
            url_builder = URLBuilder(
                fhir_api_url=self.fhir_api_url,
                graph_query=self.graph_query,
                main_resource_alias=resource_alias,
            )

            url = url_builder.compute()
            call = ApiGetter(url=url, elements=elements, token=self.token,)
            call.get_all()
            self.dataframes[resource_alias] = call.df
        self._clean_columns()
        for resource_alias, dataframe in self.dataframes.items():
            logger.debug(f"{resource_alias} dataframe builded head - \n{dataframe.to_string()}")
        # We check if there is more than 1 alias of resource
        if len(self.dataframes) > 1:
            self.main_dataframe = self._join()
        else:
            self.main_dataframe = list(self.dataframes.values())[0]

        logger.debug(
            f"Main dataframe builded head before columns selection-"
            f"\n{self.main_dataframe.to_string()}"
        )
        if not debug:
            self._select_columns()
        # TODO check where

    @timing
    def _select_columns(self):
        """Clean the final dataframe to keep only the columns of the select
        """
        final_columns = []
        for resource_alias, resource_alias_info in self.graph_query.resources_alias_info.items():
            for element in resource_alias_info.elements.get_subset_elements(goal="select"):
                final_columns.append(f"{resource_alias}:{element.col_name}")
        self.main_dataframe = self.main_dataframe[final_columns]

    @timing
    def _join(self) -> pd.DataFrame:
        """executes the joins one after the other in the order specified by the join_path function.

        Returns:
            pd.DataFrame -- dataframe containing all joined resources
        """
        list_join = join_path(self.graph_query.resources_alias_graph)
        main_alias_join = list_join[0][0]
        main_df = self.dataframes[main_alias_join]
        for alias_1, alias_2 in list_join:
            df_1 = main_df
            df_2 = self.dataframes[alias_2]
            main_df = self._join_2_df(alias_1, alias_2, df_1, df_2)
        return main_df

    @timing
    def _group_lines(self, df, col_name):
        logger.debug(f"dataframe before being grouped \n{df.to_string()}")
        if not df.empty:
            cols_group = [col_name]
            if cols_group:
                cols = df.columns.to_list()
                cols_list = [col_name for col_name in cols if col_name not in cols_group]
                dict_cols_list = {col: self._concatenate for col in cols_list}
                df = df.groupby(cols_group).agg(dict_cols_list)
                df.reset_index(inplace=True)
            logger.debug(f"dataframe after being grouped on {cols_group}:\n{df.to_string()}")
        return df

    @timing
    def _concatenate(self, column):
        result = []
        for list_cell in column:
            if isinstance(list_cell, list):
                result.extend([value for value in list_cell])
            else:
                result.append(list_cell)
        return result

    @timing
    def _join_2_df(
        self, alias_1: str, alias_2: str, df_1: pd.DataFrame, df_2: pd.DataFrame,
    ) -> pd.DataFrame:
        """Executes the join between two dataframes
        
        The join key is the id of the child resource.
        This id is contained in :
            * in the column child_alias:id of the child resource table named child_alias (e.g. parent:id for a parent dataframe)
            * in the alias_parent:searchparam_parent column of the parent resource table named alias_parent (e.g. condition:subject.reference for a condition dataframe)

        The function is in charge of finding out who is the mother resource and who is the daughter resource.

        Arguments:
            alias_1 {str} -- df_1 alias
            alias_2 {str} -- df_2 alias
            df_1 {pd.DataFrame} -- dataframe containing the elements of a resource
            df_2 {pd.DataFrame} -- dataframe containing the elements of a resource
        Returns:
            pd.DataFrame -- dataframe containing the elements of the 2 resources according to an inner join
        """  # noqa
        edge_info = self.graph_query.resources_alias_graph.edges[alias_1, alias_2]["info"]
        alias_parent = edge_info.parent
        alias_child = edge_info.child

        how = edge_info.join_how
        searchparam_parent = edge_info.searchparam_parent

        parent_on = f"{alias_parent}:join_{searchparam_parent}"
        child_on = f"{alias_child}:from_id"

        if how == "child":
            how = "right"
        elif how == "parent":
            how = "left"
        elif how != "inner":
            how = "inner"

        if alias_1 == alias_parent:
            df_merged_inner = pd.merge(
                left=df_1, right=df_2, left_on=parent_on, right_on=child_on, how=how,
            )
        else:
            df_merged_inner = pd.merge(
                left=df_2, right=df_1, left_on=parent_on, right_on=child_on, how=how,
            )
        return df_merged_inner

    @timing
    def _clean_columns(self):
        """performs preprocessing on all dataframes harvested in the dataframe attribute:
            * adds the resource type in front of an element id so that the resource id matches the references of its parent resource references
            * adds the table alias as a prefix to each column name
        """  # noqa
        for resource_alias, df in self.dataframes.items():
            resource_type = self.graph_query.resources_alias_info[resource_alias].resource_type

            df = df.pipe(_add_resource_type_to_id, resource_type=resource_type)

            self.dataframes[resource_alias] = df.add_prefix(f"{resource_alias}:")


def _add_resource_type_to_id(df, resource_type: str):
    # add assert to check that there is only one id in list
    df["from_id"] = f"{resource_type}/" + df["from_id"]
    return df



from fhir2dataset.query import Query  # noqa
from fhir2dataset.fhirrules_getter import FHIRRules  # noqa
from fhir2dataset.parser import FHIR2DatasetParser  # noqa
from fhir2dataset.timer import timing  # noqa



import re
from collections import defaultdict, OrderedDict
from collections.abc import Iterable

PREFIX = ["eq", "ne", "gt", "lt", "ge", "le", "sa", "eb", "ap"]


class FHIR2DatasetParser:
    def __init__(self):
        self.CLAUSES = OrderedDict()
        self.__split_mask_clauses = ""

        self.__init_clauses()
        self.__init_split_mask()
        self.__reset_config()

    def parse(self, SQL_string: str):
        self.__reset_config()
        SQL_string = SQL_string.replace("\n", " ")

        _RE_COMBINE_WHITESPACE = re.compile(r"\s+")
        SQL_string = _RE_COMBINE_WHITESPACE.sub(" ", SQL_string).strip()

        items = re.split(self.__split_mask_clauses, SQL_string)

        if "" in items:
            items.remove("")
        for clause in self.CLAUSES.keys():
            for idx, item in enumerate(items):
                if item.upper() == clause:
                    self.CLAUSES[item.upper()](items[idx + 1])

        self.__create_config()
        return self.config

    def __create_config(self):
        self.config["from"] = dict(self.__from)
        self.config["select"] = dict(self.__select)
        self.config["where"] = dict(self.__where)
        self.config["join"]["inner"] = dict(self.__inner_join)
        self.config["join"]["child"] = dict(self.__child_join)
        self.config["join"]["parent"] = dict(self.__parent_join)

        self.config["join"] = {key: value for key, value in self.config["join"].items() if value}
        self.config = {key: value for key, value in self.config.items() if value}

    def __select_parser(self, string):
        item_parsed = re.split(self.__split_mask_select, string)
        for item in item_parsed:
            alias, select_rule = re.split(r"\.", item, 1)
            assert alias in self.__from.keys()
            self.__select[alias].append(select_rule)

    def __from_parser(self, string):
        item_parsed = re.split(self.__split_mask_from, string)
        if len(item_parsed) == 2:
            resource_type = item_parsed[0]
            alias = item_parsed[1]
            self.__from[alias] = resource_type
        elif len(item_parsed) == 1:
            resource_type = item_parsed[0]
            alias = item_parsed[0]
            self.__from[alias] = resource_type
        else:
            raise ValueError

    def __inner_join_parser(self, string):
        alias_parent, searchparam_parent, alias_child = self.__join_parser(string)
        self.__inner_join[alias_parent][searchparam_parent] = alias_child

    def __child_join_parser(self, string):
        alias_parent, searchparam_parent, alias_child = self.__join_parser(string)
        self.__child_join[alias_parent][searchparam_parent] = alias_child

    def __parent_join_parser(self, string):
        alias_parent, searchparam_parent, alias_child = self.__join_parser(string)
        self.__parent_join[alias_parent][searchparam_parent] = alias_child

    def __join_parser(self, string):
        item_parsed = re.split(self.__split_mask_from_join, string)
        assert len(item_parsed) == 2
        self.__from_parser(item_parsed[0])

        join_conditions = re.split(self.__split_mask_join, item_parsed[1])
        assert len(join_conditions) == 2

        alias_parent = None
        searchparam_parent = None
        alias_child = None
        for join_condition in join_conditions:
            alias, join_rule = re.split(r"\.", join_condition, 1)
            assert alias in self.__from.keys(), (
                f"The {alias} alias used in the {join_condition} join must have been referenced "
                "behind a FROM, INNER JOIN, PARENT JOIN or CHILD JOIN. "
                f"The only aliases referenced are here: {self.__from}"
            )
            if join_rule.strip() == "id":
                if alias_child:
                    raise ValueError
                alias_child = alias
            else:
                if alias_parent or searchparam_parent:
                    raise ValueError
                alias_parent = alias
                searchparam_parent = join_rule

        return alias_parent, searchparam_parent, alias_child

    def __where_parser(self, string):
        item_parsed = re.split(self.__split_mask_where, string)
        for where_condition in item_parsed:
            condition_parsed = re.split(self.__split_mask_where_condition, where_condition)
            assert len(condition_parsed) == 3
            alias, where_rule = re.split(r"\.", condition_parsed[0], 1)
            value = condition_parsed[2]
            if condition_parsed[1] in PREFIX:
                prefix = condition_parsed[1]
                self.__where[alias][where_rule] = {prefix: value}
            else:
                self.__where[alias][where_rule] = value

    def __init_clauses(self):
        self.CLAUSES = OrderedDict(
            {
                "FROM": self.__from_parser,
                "INNER JOIN": self.__inner_join_parser,
                "CHILD JOIN": self.__child_join_parser,
                "PARENT JOIN": self.__parent_join_parser,
                "SELECT": self.__select_parser,
                "WHERE": self.__where_parser,
            }
        )

    def __init_split_mask(self):
        self.__split_mask_clauses = self.__create_mask(
            self.CLAUSES.keys(), can_start_sentence=True, keep_separators=True
        )
        self.__split_mask_from = self.__create_mask("AS")
        self.__split_mask_select = self.__create_mask(",", optional_space_before=True)
        self.__split_mask_from_join = self.__create_mask("ON")
        self.__split_mask_join = self.__create_mask("=")
        self.__split_mask_where = self.__create_mask("AND")
        self.__split_mask_where_condition = self.__create_mask(PREFIX + ["="], keep_separators=True)

    def __create_mask(
        self,
        separators,
        keep_separators=False,
        case_insensitive: bool = True,
        can_start_sentence: bool = False,
        optional_space_before: bool = False,
        optional_space_after: bool = False,
    ):
        assert not (
            can_start_sentence and optional_space_before
        ), "Can't set to true both can_start_sentence and optional_space_before optional arguments"
        if isinstance(separators, str):
            mask = separators
        elif isinstance(separators, Iterable):
            mask = "|".join(separators)
        else:
            raise ValueError

        if keep_separators:
            mask = fr"({mask})"

        if can_start_sentence:
            mask = fr"(?:\s|^){mask}"
        elif optional_space_before:
            mask = fr"\s?{mask}"
        else:
            mask = fr"\s{mask}"

        if optional_space_after:
            mask = fr"{mask}\s?"
        else:
            mask = fr"{mask}\s"

        if case_insensitive:
            mask = fr"{mask}(?i)"

        return mask

    def __reset_config(self):
        self.__select = defaultdict(list)
        self.__from = defaultdict(str)
        self.__inner_join = defaultdict(dict)
        self.__child_join = defaultdict(dict)
        self.__parent_join = defaultdict(dict)
        self.__where = defaultdict(dict)
        self.config = {
            "from": None,
            "join": dict(),
            "where": None,
            "select": None,
        }



import os
import json
import logging
from functools import lru_cache
from dataclasses import asdict
from dacite import from_dict

from fhir2dataset.timer import timing
from fhir2dataset.fhirpath import multiple_search_dict
from fhir2dataset.data_class import SearchParameters, SearchParameter, Elements, Element

logger = logging.getLogger(__name__)

DEFAULT_METADATA_DIR = "metadata"

MAPPING_SEARCHPARAMS = {
    "code": lambda search_param, element_value: setattr(
        search_param, "code", next(iter(element_value), None)
    ),
    "expression": lambda search_param, element_value: setattr(
        search_param, "fhirpath", next(iter(element_value), None)
    ),
    "base": lambda search_param, element_value: setattr(
        search_param, "resource_types", element_value
    ),
}


class FHIRRules:
    """Class storing rules specific to the FHIR syntax and/or the FHIR API used

    Attributes:
        possible_references {dict} -- dictionary storing searchparameters that can be used from a resource after an _include or _revinclude parameter
        searchparam_to_element {dict} -- dictionary storing for each resource the fhirpath corresponding to each searchparameter (e.g. {'Organization': {'address-postalcode': 'address.postalCode'}})
        capabilitystatement {dict} -- an instance json of a CapabilityStatement resource
        searchparameters {dict} -- an instance json of a SearchParameters resource
    """  # noqa

    @timing
    def __init__(
        self,
        path: str = None,
        fhir_api_url: str = None,
        searchparameters_filename: str = "SearchParameters.json",
    ):
        """
        Keyword Arguments:
            fhir_api_url {str} -- The Service Base URL (e.g. http://hapi.fhir.org/baseR4/) (default: {None})
            path {str} -- path to the folder containing capabilitystatement_filename and searchparameters_filename files (default: {None})
            searchparameters_filename {str} -- filename of a json that contains a resource of type SearchParameters  (default: {"SearchParameters.json"})
        """  # noqa
        self.fhir_api_url = fhir_api_url
        if not path:
            path = os.path.join(os.path.dirname(__file__), DEFAULT_METADATA_DIR)
        self.path = path
        self.searchparameters_filename = searchparameters_filename
        self.searchparameters = self._get_searchparameters()

    @timing
    @lru_cache(maxsize=200)
    def searchparam_to_fhirpath(self, search_param: str, resource_type: str = "all"):
        """retrieves the fhirpath that allows to retrieve the element that is the object of a searchparam in a json instance (after the 'resource' key) of a resource of a certain type

        Arguments:
            resource_type {str} -- name of a resource type (e.g. 'Organization')
            search_param {str} -- name of a searchparam of this resource type (e.g. 'address-postalcode')

        Returns:
            str -- the fhirpath for retrieving the element that is the subject of the searchparam (e.g. 'address.postalCode')
        """  # noqa
        try:
            return self.searchparameters.searchparam_to_fhirpath(search_param, resource_type)
        except KeyError:
            logger.warning(f"The searchparam '{search_param}' doesn't exist in the rules")
            return None

    @timing
    def _get_searchparameters(self) -> dict:
        """builds an instance of SearchParameters storing all the possible searchparameters (instance of SearchParameter) whose information comes from a bundle composed only of FHIR resources of type SearchParameter according to the following process:
        1. Retrieves information of interest in each instance of the bundle. For each instance of the bundle are retrieved: the code, the expression (=fhirpath) and the base (=the types of resources on which these fhirpaths can be applied)
        2. With these 3 pieces of information a SearchParameter object is created
        3. SearchParameter objects are stored in a SearchParameters object.

        Returns:
            SearchParameters -- instance described above
        """  # noqa
        bundle = self._get_from_file(self.path, self.searchparameters_filename)
        elements_empty = Elements(
            elements=[
                Element(col_name="code", fhirpath="SearchParameter.code",),
                Element(col_name="expression", fhirpath="SearchParameter.expression",),
                Element(col_name="base", fhirpath="SearchParameter.base",),
            ]
        )
        elements_empty = asdict(elements_empty)
        resources = [resource["resource"] for resource in bundle["entry"]]
        raw_list_elements = multiple_search_dict(resources, elements_empty)
        search_parameters = SearchParameters()
        for idx, raw_elements in enumerate(raw_list_elements):
            elements = from_dict(data_class=Elements, data=raw_elements)
            search_param = SearchParameter()
            for element in elements.elements:
                MAPPING_SEARCHPARAMS[element.col_name](search_param, element.value)
            if search_param.fhirpath and search_param.resource_types:
                search_parameters.add(search_param)
            else:
                logger.warning(
                    f"\nthe instance of SearchParameter named "
                    f"{search_param.code}"
                    f" has no fhirpath associated"
                )
                logger.debug(f"{resources[idx]}\n")
        return search_parameters

    @timing
    def _get_from_file(self, path: str, filename: str) -> dict:
        """Get a json (dict) from a file

        Arguments:
            path {str} -- path to the folder containing the filename
            filename {str} -- filename of the json

        Returns:
            dict -- dict containing the json
        """
        with open(os.path.join(path, filename)) as json_file:
            file_dict = json.load(json_file)
        return file_dict



import pandas as pd
import requests
import logging
from itertools import product
from typing import Type
from dataclasses import asdict
from dacite import from_dict

from fhir2dataset.timer import timing
from fhir2dataset.fhirpath import multiple_search_dict
from fhir2dataset.data_class import Elements

logger = logging.getLogger(__name__)


def concat_cell(dataset, cols_name, element):
    # column of one row, this single cell is of the same type as element.value, i.e. a list
    column = [element.value]
    dataset.append(column)
    cols_name.append(element.col_name)
    return dataset, cols_name


def concat_col(dataset, cols_name, element):
    for index, value in enumerate(element.value):
        # column of one row, this single cell is of the same type as value
        column = [value]
        dataset.append(column)
        cols_name.append(f"{element.col_name}_{index}")
    return dataset, cols_name


def concat_row(dataset, cols_name, element):
    # column of x=len(element.value) rows
    column = element.value
    dataset.append(column)
    cols_name.append(element.col_name)
    return dataset, cols_name


MAPPING_CONCAT = {"cell": concat_cell, "col": concat_col, "row": concat_row}


class Response:
    """class that contains the data retrieves from an url
    """

    def __init__(self):
        self.status_code = None
        self.results = None
        self.next_url = None


class CallApi:
    """generic class that manages the sending and receiving of a url request to a FHIR API.
    """

    @timing
    def __init__(self, url: str, token: str = None):
        self.url = url
        self.total = None
        self.auth = BearerAuth(token)

    @timing
    def get_response(self, url: str):
        """sends the request and stores the elements of the response

        Arguments:
            url {str} -- url of the request
        """
        if self.total is None:
            self.total = self._get_count(url)
            logger.info(f"there is {self.total} matching resources for {url}")
        if self.total != 0:
            # response_url contains the data returned by url
            response_url = Response()
            response = self._get_bundle_response(url)
            response_url.status_code = response.status_code
            if "entry" in response.json():
                # if no resources match the request, response_url.results = None
                logger.debug(f"status code of KeyError response:\n{response.status_code}")
                logger.debug(f"content of the KeyError response:\n{response.content}")
                response_url.results = response.json()["entry"]
                links = response.json().get("link", [])
                next_page = [l["url"] for l in links if l["relation"] == "next"]
                if len(next_page) > 0:
                    response_url.next_url = next_page[0]
                else:
                    logger.info("no next url was found")
            else:
                logger.info(f"Got a KeyError - There's no entry key in the json data we received.")
        else:
            response_url = Response()
        return response_url

    @timing
    def _get_count(self, url):
        if url[-1] == "?":
            url_number = f"{url}_summary=count"
        else:
            url_number = f"{url}&_summary=count"
        logger.info(f"Get {url_number}")
        response = requests.get(url_number, auth=self.auth)
        count = response.json().get("total")
        if count is None:
            logger.warning(f"status code of failing response:\n{response.status_code}")
            logger.warning(f"content of the failing response:\n{response.content}")
        return count

    @timing
    def _get_bundle_response(self, url):
        logger.info(f"Get {url}")
        return requests.get(url, auth=self.auth)


class BearerAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        if self.token:
            r.headers["Authorization"] = "Bearer " + self.token
        return r


class ApiGetter(CallApi):
    """class that manages the sending and receiving of a url request to a FHIR API and then transforms the answer into a tabular format

    Attributes:
        elements (Elements): instance of the FHIR2Dataset Elements class that allows to list all the elements that need to be retrieved from the bundles returned in response to the request url
        df (pd.DataFrame): dataframe containing the elements to be recovered in tabular format
    """  # noqa

    @timing
    def __init__(
        self, url: str, elements: Type[Elements], token: str = None,
    ):
        CallApi.__init__(self, url, token)
        self.elements = elements
        self.df = self._init_data()

    @timing
    def get_all(self):
        """collects all the data corresponding to the initial url request by calling the following pages
        """  # noqa
        next_url = self.url
        while next_url:
            next_url = self.get_next(next_url)

    @timing
    def get_next(self, next_url):
        """retrieves the responses contained in the following pages and stores the data in data attribute
        """  # noqa
        if next_url:
            response = self.get_response(next_url)
            next_url = self._get_data(response)
            return next_url
        else:
            logger.info("There is no more pages")
            return None

    @timing
    def _get_data(self, response: Type[Response]):
        """retrieves the necessary information from the json instance of a resource and stores it in the data attribute
        """  # noqa
        elements_empty = asdict(self.elements)
        if not response.results:
            columns = [element.col_name for element in self.elements.elements]
            df = pd.DataFrame(columns=columns)
            self.df = pd.concat([self.df, df])
            logger.info(
                "the current page doesnt have an entry keyword, therefore an empty df is created"
            )
        else:
            data_dicts = multiple_search_dict(
                [json_resource["resource"] for json_resource in response.results], elements_empty
            )
            for data_dict in data_dicts:
                elements = from_dict(data_class=Elements, data=data_dict)
                df = self._flatten_item_results(elements)
                self.df = pd.concat([self.df, df])
        return response.next_url

    @timing
    def _flatten_item_results(self, elements: Type[Elements]):
        """creates the tabular version of the elements given as input argument.
        For each element of elements, at least one column is added according to the following process.
        1. The first step is to reproduce the type of concatenation desired for each element
        If the concatenation type of the element is:
            * cell: a single column is created with a single row. The single cell is therefore of the same type of element.value, i.e. a list.
            * row: a single column is created and creates a row for each element in the element.value list.
            * col: len(element.value) column are created. Each column contains a single cell composed of an element from the element.value list.

        2. The second step is to produce the product of all possible combinations between columns.
        For example, if at the end of step 1, the table is : 
        Col_1 | Col_2 | Col_3
        pat_1 | Pete  | Ginger
        pat_1 | P.    | Ginger
              | Peter | G.

        The table resulting from step 2 will be : 
        Col_1 | Col_2 | Col_3
        pat_1 | Pete  | Ginger
        pat_1 | Pete  | G.
        pat_1 | P.    | Ginger
        pat_1 | P.    | G.
        pat_1 | Peter | Ginger
        pat_1 | Peter | G.

        Args:
            elements (fhir2dataset.Elements): instance of elements

        Returns:
            pd.DataFrame: resulting dataframe
        """  # noqa
        cols_name = []
        dataset = []

        for element in elements.elements:
            MAPPING_CONCAT[element.concat_type](dataset, cols_name, element)

        df = pd.DataFrame(list(product(*dataset)), columns=cols_name)
        return df

    def _init_data(self) -> dict:
        """generation of a dictionary whose keys correspond to the column name and the value to an empty list

        Returns:
            dict -- dictionary described above
        """  # noqa
        data = dict()
        for element in self.elements.elements:
            data[element.col_name] = []
        return pd.DataFrame(data)



"""classes representing the different types of information manipulated in FHIR2Dataset
"""
import logging
from collections import defaultdict
from pprint import pformat
from dataclasses import dataclass, field
from typing import Type, List, Optional

from fhir2dataset.visualization_tools import custom_repr


logger = logging.getLogger(__name__)


@dataclass
class SearchParameter:
    code: Optional[str] = field(default=None,)
    fhirpath: Optional[str] = field(default=None,)
    resource_types: Optional[List[str]] = field(default=None,)
    prefix: Optional[str] = field(default=None,)
    value: Optional[str] = field(default=None,)


@dataclass
class Element:
    """unit entity that can be found in a fhir instance in json format using fhirpath or in a table in the col_name column.

    An element is destined to be associated with a single value. For example, if we want to retrieve the patient id of a bundle containing 2 patients, we will have to create 2 element instances.
    """  # noqa

    col_name: str
    fhirpath: str
    goal: str = field(default="select",)
    value: Optional[list] = field(default=None,)
    concat_type: Optional[str] = field(default="cell",)
    search_parameter: Optional[SearchParameter] = field(default=None,)


@dataclass
class Elements:
    """collection allowing to group together elements
    """

    elements: List[Element] = field(default_factory=list)

    def append(self, x):
        self.elements.append(x)

    def get_subset_elements(self, goal):
        return [element for element in self.elements if element.goal == goal]


@dataclass
class ResourceAliasInfoBasic:
    alias: str
    resource_type: str
    elements: Elements() = field(default_factory=Elements())


class ResourceAliasInfo(ResourceAliasInfoBasic):
    def __repr__(self):
        return custom_repr(super().__repr__())


@dataclass
class EdgeInfo:
    parent: str
    child: str
    searchparam_parent: Optional[str] = field(default=None,)
    join_how: str = field(default="inner",)
    searchparam_prefix: Optional[dict] = field(default=None,)


class SearchParameters:
    def __init__(self, search_parameters: List[Type[SearchParameter]] = None):
        if search_parameters:
            self.items = search_parameters
        else:
            self.items = []
        self._init_data()

    def add(self, search_parameters):
        if isinstance(search_parameters, SearchParameter):
            self.items.append(search_parameters)
            self._add_data(search_parameters)
        elif isinstance(search_parameters, list):
            self.items.extend(search_parameters)
            for search_parameter in search_parameters:
                self._add_data(search_parameter)
        else:
            raise TypeError(
                f"{search_parameters} should be a list or a SearchParameter type instead of "
                f"{type(search_parameters)} type"
            )

    def searchparam_to_fhirpath(self, search_param: str, resource_type: str = "all"):
        """retrieves the fhirpath that allows to retrieve the element that is the object of a searchparam in a json instance (after the 'resource' key) of a resource of a certain type

        Arguments:
            resource_type {str} -- name of a resource type (e.g. 'Organization')
            search_param {str} -- name of a searchparam of this resource type (e.g. 'address-postalcode')

        Returns:
            str -- the fhirpath for retrieving the element that is the subject of the searchparam (e.g. 'address.postalCode')
        """  # noqa
        try:
            return self._data[search_param][resource_type]
        except KeyError:
            logger.warning(f"The searchparam '{search_param}' doesn't exist in the rules")
            return None

    def _init_data(self):
        self._data = defaultdict(lambda: defaultdict(dict))
        for search_parameter in self.items:
            self._add_data(search_parameter)

    def _add_data(self, search_parameter: Type[SearchParameter]):
        fhirpath = search_parameter.fhirpath
        data = self._data[search_parameter.code]
        for resource_type in search_parameter.resource_types:
            if resource_type not in data.keys():
                data[resource_type] = fhirpath
            else:
                raise ValueError(
                    f"the search parameter {search_parameter.code} is already recorded\n"
                    f"data already recorded: {pformat(self._data[search_parameter.code])}\n"
                    f"data given as argument: {search_parameter}\n"
                )



from functools import wraps
from time import time
import logging

logger = logging.getLogger(__name__)


def timing(f):
    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        logger.debug(f"func:{f.__qualname__}\nargs:{args},\nkargs:{kw}\ntook: {te-ts:2.6f} sec")
        return result

    return wrap
